// pic_8259.cpp

#include <iostream.h>
#include <dos.h>  // used for port I/O and memory access

#include "PIC_8259.h"
#include "mut_excl.h"    // prevents conflicting use of hardware by
                           // independent tasks.

   enum  EOI_SETTING {
       NON_SPECIFIC,
       SPECIFIC,
       AUTOMATIC};
       // these types are used in conjunction with
       // USE_AUTOMATIC_ROTATION_FOR_PRIORITY   @SEE PIC_8259.doc/3.7

   const char  Invalid_Boolean_Value[]  = "invalid boolean value";

   const char Invalid_Vector_Address[] = "invalid Vector_Address";

   const char End_Of_Init_Not_Called[] = "\"End_of_Init\" not called";

   static Master_PIC_8259* The_Master_8259;

   unsigned char PIC_8259::Masters_List_Of_Slaves = 0;

   unsigned char PIC_8259::Slaves_List_Of_Master_Level = 0;

   unsigned char PIC_8259::End_Of_Init_Has_Been_Called = FALSE;

   unsigned char PIC_8259::Special_Fully_Nested_Mode_In_Use = FALSE;

   Mutual_Exclusion Exclusive_Use_Of_8259;

   //*******************************************************************

   PIC_8259::PIC_8259 (
      unsigned char Interrupt_Control_Port_Addr_0,
      unsigned char Interrupt_Control_Port_Addr_1)
   {
      if (Interrupt_Control_Port_Addr_0 ==
              Interrupt_Control_Port_Addr_1)
      {
         abend ("invalid port addresses");
      }

      End_Of_Init_Has_Been_Called = FALSE;

      Port_0 = Interrupt_Control_Port_Addr_0;
      Port_1 = Interrupt_Control_Port_Addr_1;

   }

   void PIC_8259::Initialize_8259 (
         unsigned char Vector_Block_Value,      // range 0..31
         unsigned char Only_One_Chip,
         unsigned char Use_Level_Triggered_Mode,
         unsigned char Use_Auto_EOI,
         unsigned char Bufferred_Mode)
   {

      // check parameters:
      if (Only_One_Chip > 1)
      {
         abend (Invalid_Boolean_Value);
      }
      if (Use_Level_Triggered_Mode > 1)
      {
         abend (Invalid_Boolean_Value);
      }
      if (Use_Auto_EOI > 1)
      {
         abend (Invalid_Boolean_Value);
      }
      if (Bufferred_Mode > 1)
      {
         abend (Invalid_Boolean_Value);
      }
      if (Vector_Block_Value > MAX_VECTOR_BLOCK)
      {
         abend ("Invalid Vector_Block_Value");
      }

      Exclusive_Use_Of_8259.Save_Int_State_And_Disable_Int ();

         // ICW1 must be written:
         // specify that ICW4 is required:
      outportb (Port_0,  0x11 +
            (Use_Level_Triggered_Mode << 3) +
            (Only_One_Chip << 1));

         //..............................................................
         // ICW2 must be written:

      outportb (Port_1,  0x80 +
         (Vector_Block_Value << 3));

         // ICW3 and ICW4 are written by derived classes...


      // Do not Restore_Previous_Int_Flag yet because
      // initialization is incomplete.  Derived class will enable
      // interrupts.
   }

   void PIC_8259::End_Of_All_Initializations ()
   {
      if (Masters_List_Of_Slaves !=
          Slaves_List_Of_Master_Level)
      {
         abend ("Disagreement between master list and slave list");
      }

      End_Of_Init_Has_Been_Called = TRUE;
   }

   void PIC_8259::Disable_Interrupts
      (unsigned char The_Interrupts[],
       int           Number_Of_Interrupts)
         // specified interrupt level(s) will be disabled (others not
         // affected)
   {
      unsigned char IMR;

      char  i;

      if (! End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      if (Number_Of_Interrupts == 0)      // nothing to do
         return;

      Exclusive_Use_Of_8259.Save_Int_State_And_Disable_Int ();

      IMR = inportb (Port_1);

      for (i = 0; i < Number_Of_Interrupts; i++)

         // clear the bit corresponding to the int level:
         IMR &= (0xFF ^ 1 <<  The_Interrupts[i]);

      outportb (Port_1, IMR);

      Exclusive_Use_Of_8259.Restore_Previous_Int_Flag ();

   }

   void PIC_8259::Enable_Interrupts
      (unsigned char The_Interrupts[],
       int           Number_Of_Interrupts)
   {
      unsigned char IMR;

      char i;

      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      if (Number_Of_Interrupts == 0)      // nothing to do
         return;

      Exclusive_Use_Of_8259.Save_Int_State_And_Disable_Int ();

      IMR = inportb (Port_1);

      for (i = 0; i < Number_Of_Interrupts; i++)

         // SET the bit corresponding to the int level:
         IMR |= 1 << The_Interrupts[i];

      outportb (Port_1, IMR);

      Exclusive_Use_Of_8259.Restore_Previous_Int_Flag ();

   }

   void PIC_8259::Set_Priority_To_Be_Lowest (unsigned char The_Interrupt)
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      outportb (Port_0, 0xc0 + The_Interrupt);
   }


   void PIC_8259::Issue_Specific_EOI_Command (unsigned char The_Interrupt)
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      outportb (Port_0, 0x60 + The_Interrupt);
   }

   void PIC_8259::Issue_Non_Specific_EOI_Command ()
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      outportb (Port_0, 0x20);
   }

   void PIC_8259::Issue_Specific_EOI_Command_And_Set_Priority_To_Be_Lowest
         (unsigned char The_Interrupt)
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      outportb (Port_0, 0xe0 + The_Interrupt);
   }

   void 
      PIC_8259::Issue_Non_Specific_EOI_Command_And_Set_Priority_To_Be_Lowest()
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      outportb (Port_0, 0xa0);
   }

   void PIC_8259::Enable_Automatic_Rotation_For_AEOI ()
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      outportb (Port_0, 0x80);
   }

   void PIC_8259::Disable_Automatic_Rotation_For_AEOI ()
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      outportb (Port_0, 0x00);
   }

   void PIC_8259::Enable_Special_Mask_Mode ()
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      outportb (Port_0, 0x68);
   }

   void PIC_8259::Disable_Special_Mask_Mode ()
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      outportb (Port_0, 0x48);
   }

   void PIC_8259::Poll_8259_And_Determine_Highest_Interrupt_Request
      (unsigned char &Interrupt_Is_Requested,
       unsigned char &Interrupt_Level)
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      unsigned char Poll_Byte;

      Exclusive_Use_Of_8259.Save_Int_State_And_Disable_Int ();

      outportb (Port_0, 0x0c);

      Poll_Byte = inportb (Port_0);

      if (Poll_Byte & 0x80)   // indicates there was an interrupt
      {
         Interrupt_Is_Requested = TRUE;
         Interrupt_Level = Poll_Byte & 7;
      }
      else
         Interrupt_Is_Requested = FALSE;

      Exclusive_Use_Of_8259.Restore_Previous_Int_Flag ();

      return;

   }

   unsigned char PIC_8259::Interrupts_In_Service ()
   {

      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      outportb (Port_0, 0x0b);

      return inportb (Port_0);
   }

   unsigned char PIC_8259::Disabled_Interrupts ()
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      return inportb (Port_1);

   }

   unsigned char PIC_8259::Interrupts_Requesting_Service ()
   {
      if (! PIC_8259::End_Of_Init_Has_Been_Called)
      {
         abend (End_Of_Init_Not_Called);
      }

      outportb (Port_0, 0x0a);

      return inportb (Port_0);
   }


   //*******************************************************************
//        Master_PIC_8259:

   void Master_PIC_8259::Initialize_Master_8259 (
         unsigned char Vector_Block_Value,      // range 0..31
         unsigned char Interrupt_Levels_Which_Are_Slaves [],
         int           Number_Of_Slaves,
         unsigned char Use_Special_Fully_Nested_Mode,
         unsigned char Use_Level_Triggered_Mode,
         unsigned char Use_Auto_EOI,
         unsigned char Bufferred_Mode)

   {

      char ICW3 = 0;

      char i;

      // Exclusive_Use_Of_8259.Save_Int_State_And_Disable_Int () is
      //    invoked in Initialize_8259.

      Initialize_8259 (
            Vector_Block_Value,
            Number_Of_Slaves == 0 ? TRUE : FALSE,
            Use_Level_Triggered_Mode,
            Use_Auto_EOI,
            Bufferred_Mode);

         // ICW3 is written for cascade mode only:

      if (Number_Of_Slaves)
      {
         for (i = 0; i < Number_Of_Slaves; i++)
         {
            // verify valid interrupt level and
            // it is specified only once (unless chip is being
            //   initialized second time):
            if ((Interrupt_Levels_Which_Are_Slaves[i] > 7) ||
                ((! PIC_8259::End_Of_Init_Has_Been_Called)
                  && (1 << Interrupt_Levels_Which_Are_Slaves[i]) & ICW3 != 0))
                  {
                     abend ("invalid Interrupt_Level");
                  }

            ICW3 += 1 << Interrupt_Levels_Which_Are_Slaves[i];
         }

         outportb (Port_1, ICW3);
         Masters_List_Of_Slaves = ICW3;
      }

         // ICW4 must be written for 80X86:

         outportb (Port_1, 0x05 +
                           Use_Special_Fully_Nested_Mode << 4 +
                           Bufferred_Mode << 3 +
                           Use_Auto_EOI << 2);

         Exclusive_Use_Of_8259.Restore_Previous_Int_Flag ();

         The_Master_8259 = this;

         Special_Fully_Nested_Mode_In_Use = Use_Special_Fully_Nested_Mode;

   }

   //*******************************************************************

   // This class is for slave 8259's.

   void Slave_PIC_8259::Initialize_Slave_8259 (
         unsigned char Vector_Block_Value,      // range 0..31
         unsigned char Slave_ID,
         unsigned char Use_Level_Triggered_Mode,
         unsigned char Use_Auto_EOI,
         unsigned char Bufferred_Mode)
   {

      Masters_Interrupt_Level = Slave_ID;

      Initialize_8259 (
            Vector_Block_Value,
            FALSE,
            Use_Level_Triggered_Mode,
            Use_Auto_EOI,
            Bufferred_Mode);

         // ICW3 must be written
         outportb (Port_1, Slave_ID);

         // verify that ID is written only once:
         // (unless chip is being initialized second time)

         if ((Slave_ID > 7) ||
             ((! PIC_8259::End_Of_Init_Has_Been_Called) &&
             ((1 << Slave_ID) & PIC_8259::Slaves_List_Of_Master_Level) != 0))
            {
               abend ("invalid Interrupt_Level");
            }

         // set bit corresponding to level to "1":
         PIC_8259::Slaves_List_Of_Master_Level += 1 << Slave_ID;

         // ICW4 must be written for 80X86:

         outportb (Port_1, 0x01 +
                           Bufferred_Mode << 3 +
                           Use_Auto_EOI << 2);
         // Note that Special Fully Nested Mode is only programmed for
         // master.

         Exclusive_Use_Of_8259.Restore_Previous_Int_Flag ();

   }

   void Slave_PIC_8259::Issue_Specific_EOI_Command
      (unsigned char The_Interrupt)

   {
      Exclusive_Use_Of_8259.Save_Int_State_And_Disable_Int ();

      PIC_8259::Issue_Specific_EOI_Command(The_Interrupt);
      // this clears ISR bit for this interrupt level.

      if (Special_Fully_Nested_Mode_In_Use)
      // see if any other interrupts are being serviced:
         if (! Interrupts_In_Service())
            // and clear the master ISR if no other interrupts are being
            // serviced.
            The_Master_8259->Issue_Specific_EOI_Command
                  (Masters_Interrupt_Level);
      else
         // if not Special_Fully_Nested_Mode just clear master
         The_Master_8259->Issue_Specific_EOI_Command
               (Masters_Interrupt_Level);
         
      Exclusive_Use_Of_8259.Restore_Previous_Int_Flag ();
   }

   void Slave_PIC_8259::Issue_Non_Specific_EOI_Command ()
         // This routine will also issue a SPECIFIC_EOI_COMMAND to
         // the master if no other interrupts on this chip (the slave)
         // are requesting service.
   {
      Exclusive_Use_Of_8259.Save_Int_State_And_Disable_Int ();

      PIC_8259::Issue_Non_Specific_EOI_Command();
      // this clears ISR bit for this interrupt level.

      if (Special_Fully_Nested_Mode_In_Use)
      // see if any other interrupts are being serviced:
         if (! Interrupts_In_Service())
            // and clear the master ISR if no other interrupts are being
            // serviced.
            The_Master_8259->Issue_Non_Specific_EOI_Command();
      else
         // if not Special_Fully_Nested_Mode just clear master
         The_Master_8259->Issue_Non_Specific_EOI_Command();
         
      Exclusive_Use_Of_8259.Restore_Previous_Int_Flag ();
   }
